from sympy import exp, I, simplify, conjugate, log, sympify, Symbol, expand_log
from functools import reduce

from ._base import Group


class U1(Group):
    """U1 Group. Technically continuous, but represented
    in this module as discrete"""

    def __new__(cls):
        return super().__new__(cls, "U", 1)

    def __init__(self, *args, **kwargs):
        self._theta = Symbol(r"\theta", real=True)

    def generators(self) -> list:
        """Infinite circle group generators
        """
        return [exp(I * self._theta)]

    def _from_charge(self, q):
        return exp(I * self._theta * q)

    def product(self, *args, **kwargs) -> list:
        """Sums up all the charges"""
        return [simplify(reduce(lambda a, b: a*b, args))]

    def sym_product(self, *args, as_tuple=False, **kwargs) -> list:
        """Sums up all the charges, symbolically"""

        result = self.product(*[self._from_charge(x) for x in args], **kwargs)

        if as_tuple:
            return [(self._from_charge(x), x) for x in result]
        return result

    def conjugate(self, rep, symbolic=False, **kwargs):
        r"""Finds the conjugate representation of the U1 representation

        Examples
        =========
        >>> from liesym import U1
        >>> from sympy import *
        >>> u1 = U1()
        >>> u1.conjugate("1/6", symbolic=True)
        -1/6
        >>> print(u1.irrep_lookup("1/6"))
        exp(I*\theta/6)
        """
        if symbolic:
            cleaned_rep = sympify(rep) if isinstance(rep, str) else rep
            math_rep = self._from_charge(cleaned_rep)
            return expand_log(log(conjugate(math_rep)), force=True) / (I * self._theta)

        return conjugate(rep)

    def irrep_lookup(self, irrep):
        cleaned_rep = sympify(irrep) if isinstance(irrep, str) else irrep
        return self._from_charge(cleaned_rep)

    @staticmethod
    def dynkin_index(charge):
        """Returns the dynkin index for the abelian gauge"""
        return charge**2



import numpy as np
from sympy import S, Rational, Matrix

from liesym.algebras._backend import _LieAlgebraBackendWrapped
from liesym.algebras import A, B, C, D


def testBackendA():
    algebra = A(3)

    args = (
        algebra.rank,
        algebra.n_pos_roots,
        algebra.simple_roots,
        algebra.cartan_matrix,
        algebra.cartan_matrix.pinv(),
        algebra.omega_matrix,
        algebra.omega_matrix.pinv(),
        algebra.cocartan_matrix,
    )

    def _test_backend(
        rank,
        roots,
        simple_roots,
        cartan_matrix,
        cartan_matrix_inverse,
        omega_matrix,
        omega_matrix_inverse,
        cocartan_matrix
    ):
        assert isinstance(rank, int)
        assert isinstance(roots, int)

        assert isinstance(simple_roots, np.ndarray)
        assert simple_roots.shape == (3, 4, 2)

        assert isinstance(cartan_matrix, np.ndarray)

        assert cartan_matrix.shape == (3, 3, 2)

        assert isinstance(cartan_matrix_inverse, np.ndarray)
        assert cartan_matrix_inverse.shape == (3, 3, 2)

        assert isinstance(omega_matrix, np.ndarray)
        assert omega_matrix.shape == (3, 4, 2)

        assert isinstance(omega_matrix_inverse, np.ndarray)
        assert omega_matrix_inverse.shape == (4, 3, 2)

        assert isinstance(cocartan_matrix, np.ndarray)
        assert cocartan_matrix.shape == (3, 4, 2)

        def orb(w, s):
            assert isinstance(w, np.ndarray)
            assert isinstance(s, np.ndarray)
            return (np.array([[[1, 1], [2, 2]]]), np.array([[[2, 2], [3, 3]]]))

        _test_backend.root_system = lambda: (
            np.array([[[1, 1], [2, 2]]]), np.array([[[2, 2], [3, 3]]]))
        _test_backend.orbit = orb
        _test_backend.tensor_product_decomposition = lambda x, _: orb(
            x,  np.array([]))
        _test_backend.irrep_by_dim = lambda *_:  (
            np.array([[[1, 1], [2, 2]]]), np.array([[[2, 2], [3, 3]]]))

        return _test_backend

    obj = _LieAlgebraBackendWrapped(*args, backend=_test_backend)
    expected = [Matrix([[S.Half, S.Half]]), Matrix(
        [[Rational(2, 3), Rational(2, 3)]])]

    assert obj.root_system() == expected
    assert obj.orbit(expected[0], [1, 2, 3]) == expected
    assert obj.tensor_product_decomposition(*expected) == expected
    assert obj.get_irrep_by_dim(1, 2) == expected


def testBackendBCD():
    for cls in [B, C, D]:
        algebra = cls(3)

        args = (
            algebra.rank,
            algebra.n_pos_roots,
            algebra.simple_roots,
            algebra.cartan_matrix,
            algebra.cartan_matrix.pinv(),
            algebra.omega_matrix,
            algebra.omega_matrix.pinv(),
            algebra.cocartan_matrix,
        )

        def _test_backend(
            rank,
            roots,
            simple_roots,
            cartan_matrix,
            cartan_matrix_inverse,
            omega_matrix,
            omega_matrix_inverse,
            cocartan_matrix
        ):
            assert isinstance(rank, int)
            assert rank == 3

            assert isinstance(roots, int)

            assert isinstance(simple_roots, np.ndarray)
            assert simple_roots.shape == (3, 3, 2)

            assert isinstance(cartan_matrix, np.ndarray)
            assert cartan_matrix.shape == (3, 3, 2)

            assert isinstance(cartan_matrix_inverse, np.ndarray)
            assert cartan_matrix_inverse.shape == (3, 3, 2)

            assert isinstance(omega_matrix, np.ndarray)
            assert omega_matrix.shape == (3, 3, 2)

            assert isinstance(omega_matrix_inverse, np.ndarray)
            assert omega_matrix_inverse.shape == (3, 3, 2)

            assert isinstance(cocartan_matrix, np.ndarray)
            assert cocartan_matrix.shape == (3, 3, 2)

        _LieAlgebraBackendWrapped(*args, backend=_test_backend)



from sympy import Matrix, I, LeviCivita, sympify

from liesym import SU, SO, Sp, A, B, C, D


def test_su():
    su2 = SU(2)

    assert su2.dimension == 2
    assert su2.group == "SU"

    assert su2.generators() == [
        Matrix([
            [0, 1],
            [1, 0]]) / 2,
        Matrix([
            [0, -I],
            [I,  0]]) / 2,
        Matrix([
            [1,  0],
            [0, -1]]) / 2]

    assert su2.algebra == A(1)

    for i in range(3):
        for j in range(3):
            for k in range(3):
                assert su2.structure_constants(i, j, k) == LeviCivita(i, j, k)

    for n in range(2, 5):
        g = SU(n)
        assert g.quadratic_casimir(n) == sympify(n**2 - 1) / sympify(2 * n)


def test_so():
    so3 = SO(3)

    assert so3.dimension == 3
    assert so3.group == "SO"

    assert so3.generators() == [
        Matrix([
            [0, I, 0],
            [-I, 0, 0],
            [0, 0, 0]]),
        Matrix([
            [0, 0, I],
            [0, 0, 0],
            [-I, 0, 0]]),
        Matrix([
            [0,  0, 0],
            [0,  0, I],
            [0, -I, 0]])]

    assert so3.generators(True) == [
        (Matrix([
            [0, I, 0],
            [-I, 0, 0],
            [0, 0, 0]]), (1, 0)),
        (Matrix([
            [0, 0, I],
            [0, 0, 0],
            [-I, 0, 0]]), (2, 0)),
        (Matrix([
            [0,  0, 0],
            [0,  0, I],
            [0, -I, 0]]), (2, 1))
    ]

    assert so3.algebra == B(1)

    assert SO(4).algebra == D(2)

    for n in range(5, 7):
        g = SO(n)
        r = g.algebra.fundamental_weights[0]
        assert g.quadratic_casimir(r) == sympify(n - 1) / 2


def test_sp():
    sp4 = Sp(4)

    assert sp4.dimension == 4
    assert sp4.group == "Sp"

    assert len(sp4.generators()) == 10

    assert sp4.algebra == C(2)

    for n in range(2, 5):
        g = Sp(2 * n)
        r = g.algebra.fundamental_weights[0]
        assert g.quadratic_casimir(r) == sympify(2 * n + 1) / 2



from __future__ import annotations


from ._base import Group, LieGroup, commutator
from ._su import generalized_gell_mann, SU
from ._so import SO
from ._sp import Sp
from ._cyclic import Z
from ._u1 import U1

__all__ = [
    "generalized_gell_mann", "SU", "SO", "Sp", "Group", "LieGroup", "Z", "U1", "commutator"
]



from liesym import A, B, C, D, Basis
from sympy import Matrix, Rational


def test_A():

    A2 = A(2)

    # test subclass items
    assert A2.dimension == 3
    assert A2.n_pos_roots == 3

    assert A2.simple_roots == [
        Matrix([[1, -1, 0]]),
        Matrix([[0, 1, -1]]),
    ]

    a = A2.to_alpha(Matrix([[1, -1, 0]]), "ortho")
    assert a == Matrix([[1, 0]]) and a.basis == Basis.ALPHA

    x = Matrix([[1, -1, 0]])
    x.basis = "ortho"
    a = A2.to_omega(x)
    assert a == Matrix([[2, -1]]) and a.basis == Basis.OMEGA

    fw = A2.fundamental_weights[0]
    assert fw.basis == Basis.ORTHO
    assert A2.to_omega(fw) == Matrix([[1, 0]])

    # baseclass generated
    A3 = A(3)
    assert A3.cartan_matrix == Matrix([[2, -1, 0], [-1, 2, -1], [0, -1, 2]])
    assert A3.cocartan_matrix == Matrix(
        [[1, -1, 0, 0], [0, 1, -1, 0], [0, 0, 1, -1]])
    assert A3.omega_matrix == Matrix([[Rational(3, 4), Rational(-1, 4), Rational(-1, 4), Rational(-1, 4)], [Rational(
        1, 2), Rational(1, 2), Rational(-1, 2), Rational(-1, 2)], [Rational(1, 4), Rational(1, 4), Rational(1, 4), Rational(-3, 4)]])
    assert A3.metric_tensor == Matrix([[Rational(3, 4), Rational(1, 2), Rational(1, 4)], [
                                      Rational(1, 2), 1, Rational(1, 2)], [Rational(1, 4), Rational(1, 2), Rational(3, 4)]])
    assert A3.reflection_matricies == [
        Matrix([
            [0, 1, 0, 0],
            [1, 0, 0, 0],
            [0, 0, 1, 0],
            [0, 0, 0, 1]]),
        Matrix([
            [1, 0, 0, 0],
            [0, 0, 1, 0],
            [0, 1, 0, 0],
            [0, 0, 0, 1]]),
        Matrix([
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 0, 0, 1],
            [0, 0, 1, 0]])]
    assert A3.fundamental_weights == [
        Matrix([[Rational(3, 4), Rational(-1, 4), Rational(-1, 4), Rational(-1, 4)]]),
        Matrix([[Rational(1, 2), Rational(1, 2), Rational(-1, 2), Rational(-1, 2)]]),
        Matrix([[Rational(1, 4), Rational(1, 4), Rational(1, 4), Rational(-3, 4)]])]

    # backend
    assert A3.root_system() == [A3.to_ortho(x, "omega") for x in [
        Matrix([[1, 0, 1]]),
        Matrix([[-1, 1, 1]]),
        Matrix([[1, 1, -1]]),
        Matrix([[-1, 2, -1]]),
        Matrix([[0, -1, 2]]),
        Matrix([[2, -1, 0]]),
        Matrix([[0, 0, 0]]),
        Matrix([[0, 0, 0]]),
        Matrix([[0, 0, 0]]),
        Matrix([[-2, 1, 0]]),
        Matrix([[0, 1, -2]]),
        Matrix([[1, -2, 1]]),
        Matrix([[-1, -1, 1]]),
        Matrix([[1, -1, -1]]),
        Matrix([[-1, 0, -1]])
    ]]

    assert A3.positive_roots == [A3.to_ortho(x, "omega") for x in [
        Matrix([[1, 0, 1]]),
        Matrix([[-1, 1, 1]]),
        Matrix([[1, 1, -1]]),
        Matrix([[-1, 2, -1]]),
        Matrix([[0, -1, 2]]),
        Matrix([[2, -1, 0]]),
    ]]

    fund = Matrix([[1, 0, 0]])
    antifund = Matrix([[0, 0, 1]])
    decomp = A3.tensor_product_decomposition([fund, antifund])

    assert set([x.as_immutable() for x in decomp]) == set([
        Matrix([[1, 0, 1]]).as_immutable(),
        Matrix([[0, 0, 0]]).as_immutable(),
    ])

    assert A3.dim(fund) == 4

    adj = Matrix([[1, 0, 1]])
    assert A3.dim(adj) == 15

    assert A3.max_dynkin_digit(adj) == 2

    assert A3.get_irrep_by_dim(15) == [adj]

    assert A3.conjugate(Matrix([[1, 1, 0]])) == Matrix([[0, 1, 1]])


def test_A1():
    A1 = A(1)

    # test subclass items
    assert A1.dimension == 2
    assert A1.irrep_lookup('3') == Matrix([[2]])
    assert A1.root_system() == [A1.to_ortho(x, "omega") for x in [
        Matrix([[2]]), Matrix([[0]]), Matrix([[-2]])]]


def test_B():
    B2 = B(2)

    # test subclass items
    assert B2.dimension == 2
    assert B2.n_pos_roots == 4

    assert B2.simple_roots == [
        Matrix([[1, -1]]),
        Matrix([[0, 1]]),
    ]

    a = B2.to_alpha(Matrix([[1, -1]]), "ortho")
    assert a == Matrix([[1, 0]]) and a.basis == Basis.ALPHA

    x = Matrix([[1, -1]])
    x.basis = "ortho"
    a = B2.to_omega(x)
    assert a == Matrix([[2, -2]]) and a.basis == Basis.OMEGA

    fw = B2.fundamental_weights[0]
    assert fw.basis == Basis.ORTHO
    assert B2.to_omega(fw) == Matrix([[1, 0]])

    # baseclass generated
    B3 = B(3)
    assert B3.cartan_matrix == Matrix([[2, -1, 0], [-1, 2, -2], [0, -1, 2]])
    assert B3.cocartan_matrix == Matrix([[1, -1, 0], [0, 1, -1], [0, 0, 2]])
    assert B3.omega_matrix == Matrix(
        [[1, 0, 0], [1, 1, 0], [Rational(1, 2), Rational(1, 2), Rational(1, 2)]])
    assert B3.metric_tensor == Matrix([[1, 1, Rational(1, 2)], [1, 2, 1], [
                                      Rational(1, 2), 1, Rational(3, 4)]])
    assert B3.reflection_matricies == [
        Matrix([
            [0, 1, 0],
            [1, 0, 0],
            [0, 0, 1]]),
        Matrix([
            [1, 0, 0],
            [0, 0, 1],
            [0, 1, 0]]),
        Matrix([
            [1, 0,  0],
            [0, 1,  0],
            [0, 0, -1]])]
    assert B3.fundamental_weights == [
        Matrix([[1, 0, 0]]),
        Matrix([[1, 1, 0]]),
        Matrix([[Rational(1, 2), Rational(1, 2), Rational(1, 2)]])
    ]

    # backend
    assert B3.root_system() == [B3.to_ortho(x, "omega") for x in [
        Matrix([[0, 1, 0]]),
        Matrix([[1, -1, 2]]),
        Matrix([[-1, 0, 2]]),
        Matrix([[1, 0, 0]]),
        Matrix([[-1, 1, 0]]),
        Matrix([[1, 1, -2]]),
        Matrix([[-1, 2, -2]]),
        Matrix([[0, -1, 2]]),
        Matrix([[2, -1, 0]]),
        Matrix([[0, 0, 0]]),
        Matrix([[0, 0, 0]]),
        Matrix([[0, 0, 0]]),
        Matrix([[-2, 1, 0]]),
        Matrix([[0, 1, -2]]),
        Matrix([[1, -2, 2]]),
        Matrix([[-1, -1, 2]]),
        Matrix([[1, -1, 0]]),
        Matrix([[-1, 0, 0]]),
        Matrix([[1, 0, -2]]),
        Matrix([[-1, 1, -2]]),
        Matrix([[0, -1, 0]]),
    ]]

    assert B3.positive_roots == [B3.to_ortho(x, "omega") for x in [
        Matrix([[0, 1, 0]]),
        Matrix([[1, -1, 2]]),
        Matrix([[-1, 0, 2]]),
        Matrix([[1, 0, 0]]),
        Matrix([[-1, 1, 0]]),
        Matrix([[1, 1, -2]]),
        Matrix([[-1, 2, -2]]),
        Matrix([[0, -1, 2]]),
        Matrix([[2, -1, 0]]),
    ]]

    decomp = B3.tensor_product_decomposition(
        [Matrix([[1, 0, 0]]), Matrix([[1, 0, 0]]), Matrix([[1, 0, 0]])])
    assert sorted([tuple(x.tolist()) for x in decomp]) == sorted([
        tuple(x.tolist()) for x in [
            Matrix([[1, 0, 0]]),
            Matrix([[1, 0, 0]]),
            Matrix([[1, 0, 0]]),
            Matrix([[3, 0, 0]]),
            Matrix([[0, 0, 2]]),
            Matrix([[1, 1, 0]]),
            Matrix([[1, 1, 0]]),
        ]])

    adj = Matrix([[1, 0, 1]])
    assert B3.dim(adj) == 48

    assert B3.max_dynkin_digit(adj) == 2

    assert B3.get_irrep_by_dim(48) == [adj]


def test_C():
    C2 = C(2)

    # test subclass items
    assert C2.dimension == 2
    assert C2.n_pos_roots == 4

    assert C2.simple_roots == [
        Matrix([[1, -1]]),
        Matrix([[0, 2]]),
    ]

    a = C2.to_alpha(Matrix([[1, -1]]), "ortho")
    assert a == Matrix([[1, 0]]) and a.basis == Basis.ALPHA

    x = Matrix([[1, -1]])
    x.basis = "ortho"
    a = C2.to_omega(x)
    assert a == Matrix([[2, -1]]) and a.basis == Basis.OMEGA

    fw = C2.fundamental_weights[0]
    assert fw.basis == Basis.ORTHO
    assert C2.to_omega(fw) == Matrix([[1, 0]])

    # baseclass generated
    C3 = C(3)
    assert C3.cartan_matrix == Matrix([[2, -1, 0], [-1, 2, -1], [0, -2, 2]])
    assert C3.cocartan_matrix == Matrix([[1, -1, 0], [0, 1, -1], [0, 0, 1]])
    assert C3.omega_matrix == Matrix([[1, 0, 0], [1, 1, 0], [1, 1, 1]])
    assert C3.metric_tensor == Matrix([[Rational(1, 2), Rational(1, 2), Rational(
        1, 2)], [Rational(1, 2), 1, 1], [Rational(1, 2), 1, Rational(3, 2)]])
    assert C3.reflection_matricies == [
        Matrix([
            [0, 1, 0],
            [1, 0, 0],
            [0, 0, 1]]),
        Matrix([
            [1, 0, 0],
            [0, 0, 1],
            [0, 1, 0]]),
        Matrix([
            [1, 0,  0],
            [0, 1,  0],
            [0, 0, -1]])]
    assert C3.fundamental_weights == [
        Matrix([[1, 0, 0]]),
        Matrix([[1, 1, 0]]),
        Matrix([[1, 1, 1]])
    ]
    # backend
    assert [C3.to_omega(x) for x in C3.root_system()] == [
        Matrix([[2, 0, 0]]),
        Matrix([[0, 1, 0]]),
        Matrix([[-2, 2, 0]]),
        Matrix([[1, -1, 1]]),
        Matrix([[-1, 0, 1]]),
        Matrix([[1, 1, -1]]),
        Matrix([[-1, 2, -1]]),
        Matrix([[0, -2, 2]]),
        Matrix([[2, -1, 0]]),
        Matrix([[0, 0, 0]]),
        Matrix([[0, 0, 0]]),
        Matrix([[0, 0, 0]]),
        Matrix([[-2, 1, 0]]),
        Matrix([[0, 2, -2]]),
        Matrix([[1, -2, 1]]),
        Matrix([[-1, -1, 1]]),
        Matrix([[1, 0, -1]]),
        Matrix([[-1, 1, -1]]),
        Matrix([[2, -2, 0]]),
        Matrix([[0, -1, 0]]),
        Matrix([[-2, 0, 0]]),
    ]

    assert [C3.to_omega(x) for x in C3.positive_roots] == [
        Matrix([[2, 0, 0]]),
        Matrix([[0, 1, 0]]),
        Matrix([[-2, 2, 0]]),
        Matrix([[1, -1, 1]]),
        Matrix([[-1, 0, 1]]),
        Matrix([[1, 1, -1]]),
        Matrix([[-1, 2, -1]]),
        Matrix([[0, -2, 2]]),
        Matrix([[2, -1, 0]]),
    ]

    adj = Matrix([[1, 0, 1]])
    assert C3.dim(adj) == 70
    assert C3.max_dynkin_digit(adj) == 2
    assert C3.get_irrep_by_dim(70) == [adj]


def test_D():
    D2 = D(2)

    # test subclass items
    assert D2.dimension == 2
    assert D2.n_pos_roots == 2

    assert D2.simple_roots == [
        Matrix([[1, -1]]),
        Matrix([[1, 1]]),
    ]

    a = D2.to_alpha(Matrix([[1, -1]]), "ortho")
    assert a == Matrix([[1, 0]]) and a.basis == Basis.ALPHA

    x = Matrix([[1, -1]])
    x.basis = "ortho"
    a = D2.to_omega(x)
    assert a == Matrix([[2, 0]]) and a.basis == Basis.OMEGA

    fw = D2.fundamental_weights[0]
    assert fw.basis == Basis.ORTHO
    assert D2.to_omega(fw) == Matrix([[1, 0]])

    # baseclass generated
    D3 = D(3)
    assert D3.cartan_matrix == Matrix([[2, -1, -1], [-1, 2, 0], [-1, 0, 2]])
    assert D3.cocartan_matrix == Matrix([[1, -1, 0], [0, 1, -1], [0, 1, 1]])
    assert D3.omega_matrix == Matrix([[1, 0, 0], [Rational(1, 2), Rational(
        1, 2), Rational(-1, 2)], [Rational(1, 2), Rational(1, 2), Rational(1, 2)]])
    assert D3.metric_tensor == Matrix([[1, Rational(1, 2), Rational(1, 2)], [Rational(
        1, 2), Rational(3, 4), Rational(1, 4)], [Rational(1, 2), Rational(1, 4), Rational(3, 4)]])
    assert D3.reflection_matricies == [
        Matrix([
            [0, 1, 0],
            [1, 0, 0],
            [0, 0, 1]]),
        Matrix([
            [1, 0, 0],
            [0, 0, 1],
            [0, 1, 0]]),
        Matrix([
            [1,  0,  0],
            [0,  0, -1],
            [0, -1,  0]])]
    assert D3.fundamental_weights == [
        Matrix([[1, 0, 0]]),
        Matrix([[Rational(1, 2), Rational(1, 2), Rational(-1, 2)]]),
        Matrix([[Rational(1, 2), Rational(1, 2), Rational(1, 2)]])
    ]

    # backend
    assert [D3.to_omega(x) for x in D3.root_system()] == [
        Matrix([[0, 1, 1]]),
        Matrix([[1, -1, 1]]),
        Matrix([[1, 1, -1]]),
        Matrix([[-1, 0, 2]]),
        Matrix([[-1, 2, 0]]),
        Matrix([[2, -1, -1]]),
        Matrix([[0, 0, 0]]),
        Matrix([[0, 0, 0]]),
        Matrix([[0, 0, 0]]),
        Matrix([[-2, 1, 1]]),
        Matrix([[1, -2, 0]]),
        Matrix([[1, 0, -2]]),
        Matrix([[-1, -1, 1]]),
        Matrix([[-1, 1, -1]]),
        Matrix([[0, -1, -1]]),
    ]

    assert [D3.to_omega(x) for x in D3.positive_roots] == [
        Matrix([[0, 1, 1]]),
        Matrix([[1, -1, 1]]),
        Matrix([[1, 1, -1]]),
        Matrix([[-1, 0, 2]]),
        Matrix([[-1, 2, 0]]),
        Matrix([[2, -1, -1]]),
    ]

    adj = Matrix([[1, 0, 1]])
    assert D3.dim(adj) == 20
    assert D3.max_dynkin_digit(adj) == 2
    assert D3.get_irrep_by_dim(20) == [Matrix([[1, 1, 0]]),
                                       Matrix([[1, 0, 1]]),
                                       Matrix([[2, 0, 0]]),
                                       Matrix([[0, 3, 0]]),
                                       Matrix([[0, 0, 3]])]

    D5 = D(5)
    assert D5.conjugate(Matrix([[1, 0, 0, 0, 0]])) == Matrix([[1, 0, 0, 0, 0]])



from sympy import Symbol, pi, I, exp

from liesym import Z


def test_z():
    z5 = Z(5)

    g = z5.generators()

    assert [1] == z5.product(*g)

    assert [Symbol("Z_3")] == z5.sym_product("Z_1", "Z_2")

    assert [(Symbol("Z_2"), exp(4*I*pi/5))
            ] == z5.sym_product("Z_3", "Z_4", as_tuple=True)



from __future__ import annotations

from sympy.core import Basic
from sympy import Matrix, Symbol, sympify, ones
from typing import Tuple, Union
from copy import deepcopy
from functools import cmp_to_key
import re

from ._methods import (
    _cartan_matrix,
    _cocartan_matrix,
    _quadratic_form,
    _reflection_matricies,
    _annotate_matrix,
    Basis,
    _basis_lookup
)

from ._backend import create_backend


class NumericSymbol(Symbol):
    """Extension of Sympy symbol that allows
    latex formatting but also tracks the underlying 
    integer value. Useful for dimension representations
    of irreps"""
    def __new__(cls, dim: int, fmtted_dim: str):
        obj = super().__new__(cls, fmtted_dim)
        obj.numeric_dim = int(dim)
        return obj

    @classmethod
    def from_symbol(cls, symbol: Symbol):
        """Converts from sympy.Symbol into NumericSymbol by
        regex search for digits from latex display pattern and 
        returns a NumericSymbol. Will raise if no numeric is 
        present in the symbol.
        """
        try:
            s = symbol.__str__()
            num = re.findall(r"\d+", s)[0]
            return cls(int(num), s)
        except (IndexError, ValueError):
            raise ValueError("Could not extract numerical from sympy.Symbol")


class LieAlgebra(Basic):
    """The base class for all lie algebras. The methods and properties
    in this class are basis independent and apply in a general sense. In
    order to write down the roots as matricies and vectors, we choose a 
    representation.

    """
    def __new__(cls, series: str, rank: int):
        """
        Returns a new instance of a Sympy object

        Args:
            series (str): The series type
            rank (int): The rank of the algebra
        """
        obj = super().__new__(cls, sympify(rank))
        obj._series = series
        return obj

    def __init__(self, *args, **kwargs):
        """Used to set lazy properties
        """
        self._simple_roots = None
        self._positive_roots = None
        self._cartan_matrix = None
        self._omega_matrix = None
        self._quadratic_form = None
        self._cocartan_matrix = None
        self._reflection_matricies = None
        self._fundamental_weights = None
        self._backend = None
        self._root_system = None
        self._adjoint_casimir = None

    @property
    def series(self) -> str:
        """Algebra series type
        """
        return self._series

    @property
    def rank(self) -> int:
        """Algebra rank
        """
        return self.args[0]

    @property
    def dimension(self) -> int:
        """Algebra dimension

        Abstract
        """

    @property
    def n_pos_roots(self) -> int:
        """Total number of positive roots in the algebra

        Abstract
        """
    @property
    def n_roots(self) -> int:
        """Total number of roots in the algebra"""
        return 2 * self.n_pos_roots + self.rank

    @property
    def simple_roots(self) -> list[Matrix]:
        """Returns a list of Sympy matrix (1,dimension)
        objects representing a chosen basis of the algebra.

        Basis: Orthogonal


        This method can be overridden to choose your own basis,
        be sure to do this before any other properties are called
        as they are lazily evaluated and the simple_roots define
        the entire representation of the algebra.

        Examples
        ========

        .. code-block:: python

            from liesym import F4

            algebra = F4()
            my_simple_roots = [
                    # my basis
                ]
            algebra.simple_roots = my_simple_roots

        """
        return [_annotate_matrix(x) for x in self._simple_roots]

    @simple_roots.setter
    def simple_roots(self, val: list[Matrix]):
        """Overrides the default representation of the algebras simple_roots.
        Please ensure that roots are in Orthogonal Basis
        """
        assert len(val) == len(
            self._simple_roots), "Incorrect number of simple roots"
        self._simple_roots = val

    @property
    def cartan_matrix(self) -> Matrix:
        r"""For a given simple Lie algebra the elements $a_{ij}$ can be
        generated by

        .. math::
            a_{ji} = 2 \langle\alpha_i, \alpha_j\rangle / \langle\alpha_j, \alpha_j\rangle

        where $a_i$ is the i'th simple root and $\langle,\rangle$ is the scalar product.

        Sources:
            - https://en.wikipedia.org/wiki/Cartan_matrix
            - https://mathworld.wolfram.com/CartanMatrix.html

        Returns:
            Matrix: Cartan Matrix as a Sympy object
        """
        if self._cartan_matrix is None:
            self._cartan_matrix = _cartan_matrix(self.simple_roots)
        return self._cartan_matrix

    @property
    def cocartan_matrix(self) -> Matrix:
        """The cocartan matrix rows are generated from the coroots of 
        the algebra such that multiplication by a simple root will
        generate a row of the cartan matrix.

        Returns:
            Matrix: Cocartan Matrix as a Sympy object
        """
        if self._cocartan_matrix is None:
            self._cocartan_matrix = _cocartan_matrix(self.simple_roots)
        return self._cocartan_matrix

    @property
    def omega_matrix(self) -> Matrix:
        """The rows of the omega matrix are the fundamental weights
        of the algebra.

        Returns:
            Matrix: Omega Matrix as a Sympy object
        """
        if self._omega_matrix is None:
            self._omega_matrix = self.cocartan_matrix.pinv().T
        return self._omega_matrix

    @property
    def metric_tensor(self) -> Matrix:
        """Also known as the quadratic form, the metric tensor
        serves as the metrix for the inner product of two roots or weights
        when they are not in the orthogonal basis.

        Returns:
            Matrix: Metric Tensor as a Sympy object
        """
        if self._quadratic_form is None:
            self._quadratic_form = _quadratic_form(
                self.cartan_matrix, self.simple_roots)
        return self._quadratic_form

    @property
    def reflection_matricies(self) -> list[Matrix]:
        """Returns a list of reflection matrices built from
        rotations about each simple root.

        Returns:
            list[Matrix]: list of Sympy Matrices
        """
        if self._reflection_matricies is None:
            self._reflection_matricies = _reflection_matricies(
                self.simple_roots)
        return self._reflection_matricies

    @property
    def fundamental_weights(self) -> list[Matrix]:
        """Returns the fundamental weights of the algebra. 

        Basis: Orthogonal

        Returns:
            list[Matrix]: list of Sympy Matrices
        """
        if self._fundamental_weights is None:
            self._fundamental_weights = [
                _annotate_matrix(self.omega_matrix.row(i))
                for i in range(self.omega_matrix.rows)
            ]
        return self._fundamental_weights

    @property
    def positive_roots(self) -> list[Matrix]:
        """Returns the postive roots of the algebra. They are sorted 
        first by their distance from the highest root and then by 
        tuple ordering (convention).

        Basis: Orthogonal

        Returns:
            list[Matrix]: list of Sympy Matrices
        """
        if self._positive_roots is None:
            self._positive_roots = self.root_system()[:self.n_pos_roots]
        return self._positive_roots

    @property
    def _backend_instance(self):
        if self._backend is None:
            self._backend = create_backend(self)
        return self._backend

    def orbit(self, weight: Matrix, stabilizers=None, basis="ortho") -> list[Matrix]:
        """Returns the orbit of the weight or root by reflecting it
        a plane. A stabilizer may be passed to calculate the orbit using
        the Orbit-Stabilizer theorem.

        Basis: Ortho

        Args:
            weight (Matrix): A Matrix of shape (1, rank)
            stabilizer (Iterable of ints, optional): Per Orbit-Stabilizer
            theorem, integer iterable of simple root indexes. Defaults to None.

        Sources:
        - https://en.wikipedia.org/wiki/Coadjoint_representation#Coadjoint_orbit
        - https://en.wikipedia.org/wiki/Group_action#Orbits_and_stabilizers

        """

        weight = self.to_ortho(weight, "ortho")
        return [self.to_ortho(x, "ortho") for x in self._backend_instance.orbit(weight, stabilizers)]

    def dim_name(self, irrep: Matrix, basis="omega") -> NumericSymbol:
        r"""Returns a sympy formatted symbol for the irrep.
        This is commonly used in physics literature. Returns
        a NumericSymbol object that is a simple extension of 
        sympy.Symbol.      

        Examples
        =========
        >>> from liesym import A
        >>> from sympy import Matrix
        >>> a3 = A(3)
        >>> assert str(a3.dim_name(Matrix([[1, 1, 0]]))) == '\\bar{20}'           
        """
        irrep = self.to_omega(irrep, basis)

        dim = self.dim(irrep)
        max_dd = self.max_dynkin_digit(irrep)
        same_dim_irreps: list[Matrix] = self.get_irrep_by_dim(dim, max_dd)
        num_primes = 0
        conjugate = 0
        so8label = ""

        if len(same_dim_irreps) > 1:
            # group by index
            index_pairs = {}  # type: ignore
            for i in same_dim_irreps:
                index = self._backend_instance.index_irrep(i, dim)
                index_pairs[index] = index_pairs.get(index, []) + [i]

            groups = [sorted(dimindex, key=cmp_to_key(self._dimindexsort))
                      for dimindex in index_pairs.values()]
            positions = []
            for id1, grps in enumerate(groups):
                for id2, g in enumerate(grps):
                    if g == irrep:
                        positions.append([id1, id2])
            [num_primes, conjugate] = positions[0]
            so8label = self._is_s08(irrep)
        has_conjugate = conjugate == 1 if so8label == "" else False
        return self._dim_name_fmt(dim, has_conjugate, num_primes, so8label)

    def irrep_lookup(self, dim: Union[Symbol, str], max_dynkin_digit=5) -> Matrix:
        """Returns the irrep matrix for the dimension.

        Args:
            dim (Union[Symbol, str]): Can either be a sympy.Symbol or string.

        Raises:
            KeyError: Dim not found

        Returns:
            Matrix: Returns irrep in Omega basis


        Examples
        ========
        >>> from liesym import A
        >>> A3 = A(3)
        >>> A3.irrep_lookup(r"\\bar{4}")
        Matrix([[0, 0, 1]])
        >>> A3.irrep_lookup("4")
        Matrix([[1, 0, 0]])
        """
        if isinstance(dim, str):
            dim = Symbol(dim)

        if isinstance(dim, Symbol) and not isinstance(dim, NumericSymbol):
            dim = NumericSymbol.from_symbol(dim)

        n_dim = dim.numeric_dim

        dd = 0
        while dd < max_dynkin_digit:
            dd += 1
            for c in self.get_irrep_by_dim(n_dim, dd):
                if self.dim_name(c) == dim:
                    return c
        raise KeyError(f"Irrep {dim} not found.")

    def conjugate(self, irrep: Matrix) -> Matrix:
        """Finds the conjugate irrep. If it is the same
        as the original irrep, you have a Real Irrep, otherwise
        it's a Complex Irrep.

        Examples
        ========
        .. code-block:: python

            from liesym import A,D
            from sympy import Matrix

            SU4 = A(3)
            irrep_20 = Matrix([[0,1,1]])
            irrep_20bar = Matrix([[1,1,0]])
            assert irrep_20 == SU4.conjugate(irrep_20bar)

            SO10 = D(5)
            irrep_10 = Matrix([[1, 0, 0, 0, 0]])
            assert irrep_10 == SO10.conjugate(irrep_10)
        """
        return self.to_omega(self._backend_instance.conjugate(irrep)[0], "omega")

    def _is_s08(self, irrep):
        return ""

    def _dimindexsort(self, irrep1, irrep2):
        cong1 = self._congruency_class(irrep1)
        cong2 = self._congruency_class(irrep2)
        if isinstance(cong1, tuple):
            return 1 if cong1[-1] <= cong2[-1] else -1
        else:
            return -1 if cong1 <= cong2 else 1

    def _congruency_class(self, irrep):
        return 0

    def max_dynkin_digit(self, irrep: Matrix) -> int:
        """Returns the max Dynkin Digit for the representations"""
        pass

    def _dim_name_fmt(self, dim: int, conj=False, primes=0, sub="") -> NumericSymbol:
        if conj:
            irrep = r"\bar{" + str(dim) + "}"
        else:
            irrep = str(dim)

        if primes > 0:
            irrep += r"^{" + " ".join([r"\prime"] * primes) + r"}"

        if sub != "":
            irrep += r"_{" + str(sub) + r"}"

        return NumericSymbol(dim, irrep)

    def get_irrep_by_dim(self, dim: int, max_dd: int = 3, with_symbols=False) -> list[Union[Matrix, Tuple[Matrix, NumericSymbol]]]:
        r"""Gets all irreps by dimension and max dynkin digit. `max_dd` is . This algorithm brute forces searches by using `itertools.product`
        which can become expensive for large so searching max_dd > 3 will be 
        very expensive

        Args:
            dim (int): Dimension to query
            max_dd (int, optional): The max dynkin digit to use. Defaults to 3.
            with_symbols (bool, optional): Returns list of tuples of rep and latex fmt. Defaults to False.

        Returns:
            list[Union[Matrix, Tuple[Matrix,NumericSymbol]]]: If `with_symbols=True` will return a list of tuples.


        Examples
        =========
        >>> from liesym import A
        >>> from sympy import Matrix
        >>> a3 = A(3)
        >>> expected = a3.get_irrep_by_dim(20)
        >>> result = [
        ... Matrix([[1, 1, 0]]),
        ... Matrix([[0, 1, 1]]),
        ... Matrix([[0, 2, 0]]),
        ... Matrix([[3, 0, 0]]),
        ... Matrix([[0, 0, 3]])]
        >>> assert expected == result
        >>> a3.get_irrep_by_dim(20, with_symbols=True)
        [(Matrix([[1, 1, 0]]), \bar{20}), (Matrix([[0, 1, 1]]), 20), (Matrix([[0, 2, 0]]), 20^{\prime}), (Matrix([[3, 0, 0]]), \bar{20}^{\prime \prime}), (Matrix([[0, 0, 3]]), 20^{\prime \prime})]
        """
        backend_results: list[Matrix] = self._backend_instance.get_irrep_by_dim(
            dim, max_dd) or []
        results = [self.to_omega(x, "omega") for x in backend_results]
        if with_symbols:
            results = [(x, self.dim_name(x)) for x in results]
        return results

    def dim(self, irrep: Matrix, basis="omega") -> int:
        r"""Returns the dimension of the weight, root or irreducible representations.
        This follows Weyl's dimension formula:

        .. math::
            dim(w) = \prod_{\alpha\in\Delta^{+}} \frac{\langle \alpha, w + \rho\rangle}{\langle\alpha,\rho\rangle}

        where $\Delta^{+}$ are the positive roots and $rho$ is the sum of
        the positive roots: `[1] * rank`.

        Examples
        ========
        >>> from liesym import A
        >>> from sympy import Matrix
        >>> a2 = A(2)
        >>> assert a2.dim(Matrix([[1,0]])) == 3
        """
        basis = _basis_lookup(basis)
        _annotate_matrix(irrep, basis)
        irrep = self.to_omega(irrep)

        return sympify(self._backend_instance.dim(irrep))

    def root_system(self) -> list[Matrix]:
        """Returns the entire rootsystem of the algebra. This
        includes the positive, negative and zeros of the algebra.

        Basis: Orthogonal

        Returns:
            list[Matrix]: list of ordered roots.
        """
        if self._root_system is None:
            self._root_system = [self.to_ortho(
                x, basis="omega") for x in self._backend_instance.root_system()]
        return self._root_system

    def tensor_product_decomposition(self, weights: list[Matrix], basis="omega", **_) -> list[Matrix]:
        """Returns the tensor product between irreducible representations
        as a the tensor sum of the irreducible representations of their
        highest weights. This algorithm is based on Klimky's formula.


        Args:
            weights (list[Matrix]): A list of fundamental weights to take the tensor product between
            basis (str, Optional): Basis of incoming weights. If not set, will implicitly set. Defaults to 'omega'.

        Returns:
            list[Matrix]: list of weights decomposed from the tensor product. Basis: Omega


        Examples
        =========
        >>> from liesym import A
        >>> from sympy import Matrix
        >>> a2 = A(2)
        >>> results = a2.tensor_product_decomposition([Matrix([[1,0]]), Matrix([[1,0]])])
        >>> print(results)
        [Matrix([[0, 1]]), Matrix([[2, 0]])]


        """
        weights = [self.to_omega(x, basis) for x in weights]
        w = deepcopy(weights)
        i = w.pop()
        j = w.pop()

        decomp = self._backend_instance.tensor_product_decomposition(i, j)

        while len(w) > 0:
            j = w.pop()
            results = []
            for i in decomp:
                # i,j reversed because pop takes from -1 index
                results += self._backend_instance.tensor_product_decomposition(
                    j, i)
            decomp = results
        return [self.to_omega(x, "omega") for x in decomp]

    def to_ortho(self, x: Matrix, basis=None) -> Matrix:
        """Rotates to orthogonal basis

        Args:
            x (Matrix): Matrix to be rotated
            basis (optional): If `basis` attribute is not set on `x` define it here. Defaults to None.

        Raises:
            ValueError: If no `x.basis` is set and None is passed to `basis` kwarg. 

        Returns:
            Matrix: Matrix in orthogonal basis.
        """
        basis = _basis_lookup(basis)
        _annotate_matrix(x, basis)

        if x.basis is Basis.ORTHO:
            r = x
        elif x.basis is Basis.OMEGA:
            r = x * self.omega_matrix
        elif x.basis is Basis.ALPHA:
            r = x * self.cartan_matrix * self.omega_matrix
        else:
            raise ValueError(
                "Basis arg cannot be None if attribute `basis` has not been set on Matrix.")

        r.basis = Basis.ORTHO
        return r

    def to_omega(self, x: Matrix, basis=None) -> Matrix:
        """Rotates to omega basis

        Args:
            x (Matrix): Matrix to be rotated
            basis (optional): If `basis` attribute is not set on `x` define it here. Defaults to None.

        Raises:
            ValueError: If no `x.basis` is set and None is passed to `basis` kwarg. 

        Returns:
            Matrix: Matrix in omega basis.
        """
        basis = _basis_lookup(basis)
        _annotate_matrix(x, basis)

        if x.basis is Basis.OMEGA:
            r = x
        elif x.basis is Basis.ORTHO:
            r = x * self.omega_matrix.pinv()
        elif x.basis is Basis.ALPHA:
            r = x * self.cartan_matrix
        else:
            raise ValueError(
                "Basis arg cannot be None if attribute `basis` has not been set on Matrix.")

        r.basis = Basis.OMEGA
        return r

    def to_alpha(self, x: Matrix, basis=None) -> Matrix:
        """Rotates to alpha basis

        Args:
            x (Matrix): Matrix to be rotated
            basis (optional): If `basis` attribute is not set on `x` define it here. Defaults to None.

        Raises:
            ValueError: If no `x.basis` is set and None is passed to `basis` kwarg. 

        Returns:
            Matrix: Matrix in alpha basis.
        """
        basis = _basis_lookup(basis)
        _annotate_matrix(x, basis)

        if x.basis is Basis.ALPHA:
            r = x
        elif x.basis is Basis.ORTHO:
            r = x * self.omega_matrix.pinv() * self.cartan_matrix.pinv()
        elif x.basis is Basis.OMEGA:
            r = x * self.cartan_matrix.pinv()
        else:
            raise ValueError(
                "Basis arg cannot be None if attribute `basis` has not been set on Matrix.")

        r.basis = Basis.ALPHA
        return r

    def _scalar_product(self, irrep1, irrep2, basis="ortho"):
        """Scalar product between two irreps."""
        irrep1 = self.to_ortho(irrep1, basis=basis)
        irrep2 = self.to_ortho(irrep2, basis=basis)

        return irrep1.dot(irrep2)

    def quadratic_casimir(self, irrep: Matrix = None, basis="omega", **kwargs) -> Basic:
        r"""Returns the quadratic casimir for an arbitrary irreducible
        representation, $R$.

        .. math::

            \frac{1}{2}<R | R + \rho> 

        Args:
            irrep (Matrix): Irreducible representation 
            basis (str, optional): Basis of irrep. Defaults to "omega".

        Returns:
            Basic: Rational number
        """
        if irrep is None:
            # return sympify(1)
            irrep = self.adjoint_rep
            basis = irrep.basis
        irrep = self.to_omega(irrep, basis=basis)
        rho = 2 * ones(1, self.rank)
        return self._scalar_product(irrep, irrep + rho, basis="omega") / 2

    def dynkin_index(self, irrep: Matrix = None, basis="omega") -> Basic:
        r"""Calculates the dynkin index of an arbitrary irrep. This
        is also called in literature second-order index, $I_2(R)$. If no irrep 
        is passed, basis and rep default to the algebra's adjoint rep.
        For arbitrary irrep, $R$, the dynkin index, $I_2(R)$ can be calculated 
        as:

        .. math::
            I_2(R) = \frac{Dim(R)}{Dim(adj)}c(R)

        where the $dim$ is the dimension formula, $adj$ is the adjoint irrep
        and $c$ is the quadratic casimir (`Liealgebra.casimir`).
        """
        irrep = irrep or self.adjoint_rep
        basis = basis if irrep else self.adjoint_rep.basis
        d_r = self.dim(irrep, basis=basis)
        d_a = self.dim(self.adjoint_rep)
        c_r = self.quadratic_casimir(irrep, basis=basis)
        return d_r * c_r / d_a

    @property
    def adjoint_rep(self) -> Matrix:
        """Returns the adjoint irrep. This is the highest weight of the 
        root system."""
        return self.positive_roots[0]



from .liesym import (
    LieAlgebraBackend as _LieAlgebraBackend,
)
from .groups import *
from .algebras import *
import os

# C++ backend for sympy. Symbolic speed ups
os.environ["USE_SYMENGINE"] = "1"

IS_SYMENGINE = (os.environ.get("USE_SYMENGINE") == "1")
# Compiled extension imports


__all__ = [
    "A", "B", "C", "D", "F4", "G2", "E", "LieAlgebra", "NumericSymbol",
    "SU", "generalized_gell_mann", "SO", "Sp", "Group", "LieGroup", "Z"
]



from __future__ import annotations

from sympy import Matrix, zeros

from ._base import LieGroup
from ..algebras import C


def _iachello_basis(dim):
    """Basis noted in Francesco Iachello's text.
    """
    n = dim / 2

    def E(i, j):
        mat = zeros(dim)
        mat[i, j] = 1
        return mat

    for k in range(n):
        yield E(k, n+k)
        yield E(k+n, k)

    for k in range(n):
        for m in range(n):
            yield E(k, m) - E(n+m, n+k)

    for m in range(n):
        for k in range(m):
            yield E(k, n+m) + E(m, n+k)
            yield E(n+k, m) + E(n+m, k)


class Sp(LieGroup):
    """The Symplectic Group
    """

    def __new__(cls, dim: int):
        if dim % 2 != 0:
            raise NotImplementedError("Sp is not defined for odd dimensions.")
        return super().__new__(cls, "Sp", dim)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._algebra = C(self.dimension / 2)

    def generators(self) -> list['Matrix']:
        """Generators for Sp(2N). There are a lot of possible choices, so 
        we choose one based on existing literature.

        Returns:
            list[Matrix]: list of (mathematical) generators in Iachello's basis.

        Sources:
            - Iachello, F (2006). Lie algebras and applications. ISBN 978-3-540-36236-4.
        """
        return list(_iachello_basis(self.dimension))



from __future__ import annotations

from sympy import zeros, Matrix, eye, sqrt
from enum import Enum


class Basis(Enum):
    """In literature there are several different basis representations
    used for convenience of calculations.

    ORTHO: The orthogonal basis which shows non-orthogonality between the simple roots.
    OMEGA: The omega basis, also known as the dynkin basis, is the basis of the fundamental weights.
    ALPHA: The alpha basis is the basis of the simple roots.
    """
    ORTHO = 0
    OMEGA = 1
    ALPHA = 2
    UNDEF = None


def _basis_lookup(x):
    if isinstance(x, Basis):
        return x
    if isinstance(x, str):
        if x.lower() == "ortho":
            return Basis.ORTHO
        if x.lower() == "omega":
            return Basis.OMEGA
        if x.lower() == "alpha":
            return Basis.ALPHA
        raise ValueError(
            "Unnsupported basis, string choices are 'ortho','alpha','omega'")
    if x is None:
        return Basis.UNDEF

    raise ValueError(
        "Unnsupported basis, string choices are 'ortho','alpha','omega'")


def _annotate_matrix(M, basis=Basis.ORTHO):
    if getattr(M, "basis", None) is None:
        proper_basis = _basis_lookup(basis)
        setattr(M, "basis", proper_basis)

    if not isinstance(M.basis, Basis):
        proper_basis = _basis_lookup(M.basis)
        setattr(M, "basis", proper_basis)

    return M


def _cartan_matrix(simple_roots: list[Matrix]) -> Matrix:
    rank = len(simple_roots)
    cartan_matrix = zeros(rank, rank)
    for i, sr_i in enumerate(simple_roots):
        for j, sr_j in enumerate(simple_roots):
            cartan_matrix[j, i] = 2 * sr_i.dot(sr_j) / sr_i.dot(sr_i)
    return cartan_matrix


def _cocartan_matrix(simple_roots: list[Matrix]) -> Matrix:
    return Matrix([2 * x / x.dot(x) for x in simple_roots])


def _quadratic_form(cartan_matrix: Matrix, simple_roots: list[Matrix]) -> Matrix:
    rank = len(simple_roots)
    quadratic_form = zeros(rank, rank)

    # normalized constant
    n_constant = sqrt(2 / max(x.dot(x) for x in simple_roots))
    for i in range(rank):
        root = n_constant * simple_roots[i]
        quadratic_form[i, i] = root.dot(root) / 2

    return cartan_matrix.pinv() * quadratic_form


def _reflection_matricies(simple_roots: list[Matrix]) -> list[Matrix]:
    def reflection_matrix(v): return (
        eye(len(v)) - 2 * v.T * v / v.dot(v)).as_immutable()
    return [reflection_matrix(x) for x in simple_roots]



# from __future__ import annotations
from functools import reduce

from sympy import exp, I, pi, Symbol, conjugate

from ._base import Group


class Z(Group):
    """Cyclic Group
    """

    def __new__(cls, dim: int):
        return super().__new__(cls, "Z", dim)

    def __init__(self, *args, **kwargs):
        self._lookups = {Symbol(f"Z_{idx}"): x for idx,
                         x in enumerate(self.generators())}

    def generators(self, indexed=False) -> list:
        """The basis for the generators in the cyclic group are 
        in the exponential imaginary basis.

        Args:
            indexed (bool, Optional): Returns tuple of named generators. Defaults to False.

        Examples
        =========
        >>> from liesym import Z
        >>> from sympy import I, pi
        >>> Z(3).generators()
        [1, exp(2*I*pi/3), exp(-2*I*pi/3)]
        >>> Z(3).generators(indexed=True)
        [(Z_0, 1), (Z_1, exp(2*I*pi/3)), (Z_2, exp(-2*I*pi/3))]
        """
        d = self.dimension

        gens = [exp(2 * pi * I * x / d) for x in range(d)]
        if indexed:
            return [(Symbol(f"Z_{idx}"), x) for idx, x in enumerate(gens)]
        return gens

    def product(self, *args, **kwargs) -> list:
        """Product of the Cyclic representations

        Returns:
            list: To keep types standard, returns a list of length 1

        Examples
        =========
        >>> from liesym import Z
        >>> z5 = Z(5)
        >>> g = z5.generators()
        >>> z5.product(g[0], g[1], g[2], g[3], g[4])
        [1]
        """
        return [reduce(lambda a, b: a*b, args)]

    def sym_product(self, *args, as_tuple=False, **kwargs) -> list:
        """Will take the product symbolically of Cyclic representations
        of the form `lambda x: f`Z_{x}`

        Returns:
            list: To keep types standard, returns a list of length 1

        Examples
        =========
        >>> from liesym import Z
        >>> z5 = Z(5)
        >>> z5.sym_product("Z_1", "Z_2")
        [Z_3]
        >>> z5.sym_product("Z_3", "Z_4", as_tuple=True)
        [(Z_2, exp(4*I*pi/5))]
        """

        cleaned_args = [Symbol(x) if isinstance(x, str) else x for x in args]

        result = self.product(*[self._lookups[x] for x in cleaned_args])

        for k, v in self._lookups.items():
            if v == result[0]:
                if as_tuple:
                    return [(k, v)]
                return [k]
        raise ValueError("Malformed cyclic product")

    def conjugate(self, rep, symbolic=False):
        """Finds the conjugate representation of the cyclic representation

        Examples
        =========
        >>> from liesym import Z
        >>> from sympy import sympify
        >>> z3 = Z(3)
        >>> g = z3.generators()
        >>> z3.conjugate(g[0])
        1
        >>> assert z3.conjugate("Z_1", symbolic=True) == sympify("Z_2")
        """
        if symbolic:
            cleaned_rep = Symbol(rep) if isinstance(rep, str) else rep
            math_rep = self._lookups[cleaned_rep]
            conj_rep = conjugate(math_rep)
            return [k for k, v in self._lookups.items() if v == conj_rep][0]

        if len([k for k, v in self._lookups.items() if v == rep]) == 0:
            raise KeyError("Rep not in cyclic group.")
        return conjugate(rep)

    def irrep_lookup(self, irrep):
        """Returns the symbol of irrep

        Examples
        =========
        >>> from liesym import Z
        >>> from sympy import sympify
        >>> z3 = Z(3)
        >>> z3.irrep_lookup("Z_1")
        exp(2*I*pi/3)
        """
        cleaned_rep = Symbol(irrep) if isinstance(irrep, str) else irrep
        return self._lookups[cleaned_rep]



from setuptools import find_packages, setup
from setuptools.command.install import install as InstallCommand
from os import path
import subprocess
import sys

setup_requires = [
    'setuptools-rust>=0.11.6',
    'twine>=3.4.2',
    "pytest-runner"
]
try:
    from setuptools_rust import RustExtension, Binding
except ModuleNotFoundError:
    subprocess.check_call(
        [sys.executable, "-m", "pip", "install"] + setup_requires)
    from setuptools_rust import RustExtension, Binding

install_requires = [
    'numpy>=1.16',
    'sympy>=1.5',
    'symengine>=0.7.2'
]

dev_install_requires = install_requires + [
    "autopep8>=1.5.6",
    "jupyter>=1.0.0",
]

doc_install_requires = install_requires + [
    "Sphinx>=4.1.1",
    "groundwork-sphinx-theme>=1.1.1",
    "numpydoc>=1.1.0",
    "pydata_sphinx_theme>=0.5.2",
    "sphinx-math-dollar>=1.2"
]

test_requires = install_requires + [
    "pytest>=6.2.3",
]


class InstallDocs(InstallCommand):
    """ Customized setuptools install command which uses pip. """

    def run(self, *args, **kwargs):
        subprocess.check_call(
            [sys.executable, "-m", "pip", "install"] + doc_install_requires)


class InstallDev(InstallCommand):
    """ Customized setuptools install command which uses pip. """

    def run(self, *args, **kwargs):
        subprocess.check_call(
            [sys.executable, "-m", "pip", "install"] + dev_install_requires)


class InstallTest(InstallCommand):
    """ Customized setuptools install command which uses pip. """

    def run(self, *args, **kwargs):
        subprocess.check_call(
            [sys.executable, "-m", "pip", "install"] + test_requires)


class Clean(InstallCommand):
    """Cleans local directory"""
    targets = [
        "target/",
        "liesym.egg-info/",
        "build/",
        ".eggs/"
    ]

    def run(self, *args, **kwargs):
        import shutil as sh
        for t in self.targets:
            sh.rmtree(t)


this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='liesym',
    version='0.6.0',
    description='Sympy Lie Algebra extensions powered by rust.',
    rust_extensions=[RustExtension(
        'liesym.liesym',
        './Cargo.toml',
        binding=Binding.PyO3,
        debug=False
    )],
    install_requires=install_requires,
    setup_requires=setup_requires,
    python_requires=">=3.7",
    packages=find_packages(),
    test_requires=test_requires,
    zip_safe=False,
    include_package_data=True,
    author="Nathan Papapietro <npapapietro95@gmail.com>",
    url="https://github.com/npapapietro/liesym",
    classifiers=[
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Physics",
        "Programming Language :: Rust",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    long_description=long_description,
    long_description_content_type='text/markdown',
    cmdclass={
        "setupdev": InstallDev,
        "setupdocs": InstallDocs,
        "setuptests": InstallTest,
        "clean": Clean
    }
)



from __future__ import annotations

from typing import Generator
from sympy import I, Basic
from sympy.core.backend import zeros, sqrt

from ._base import LieGroup
from ..algebras import A


def generalized_gell_mann(dimension: int) -> Generator:
    """Generator of generalized Gell-Mann Matrices. Note order may
    be different than usual because we generate symmetric, then antisymmetric, 
    then diagonal.

    Args:
        dimension (int): Dimension of group

    Yields:
        Generator[Matrix]: Python Generator of (mathematical) generators

    Sources:
        - https://mathworld.wolfram.com/GeneralizedGell-MannMatrix.html

    """
    def eij(dim):
        def _(i, j):
            mat = zeros(dim)
            mat[i, j] = 1
            return mat
        return _
    E = eij(dimension)

    # Symmetric
    for k in range(dimension):
        for j in range(k):
            yield E(k, j) + E(j, k)

    # AntiSymmetric
    for k in range(dimension):
        for j in range(k):
            yield I * (E(k, j) - E(j, k))

    # Diagonal
    for l in range(dimension - 1):
        coeff = sqrt(2) / sqrt((l+1) * (l + 2))
        sum_term = sum([E(j, j) for j in range(l+1)], zeros(dimension))
        yield coeff * (sum_term - (l+1) * E(l+1, l+1))


class SU(LieGroup):
    """The Special Unitary Group
    """

    def __new__(cls, dim: int):
        if dim < 2:
            raise NotImplementedError("SU(1)==U(1) is not implemented yet.")
        return super().__new__(cls, "SU", dim)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._algebra = A(self.dimension - 1)
        self._generators = None

    def generators(self):
        r"""Generators for SU(N).  Based off the generalized Gell-Mann matrices $\lambda_a$
        the generators, $T_a$ are

        .. math::

            T_a = \frac{\lambda_a}{2}            
        """
        if self._generators is None:
            # typical normalization
            self._generators = [
                x / 2 for x in generalized_gell_mann(self.dimension)]
        return self._generators



from __future__ import annotations

from typing import Tuple, Union
from sympy import Symbol, Basic, I, Basic, trace, sympify, Matrix
from sympy.core.backend import Matrix as _CMatrix
from sympy.tensor.array.dense_ndim_array import MutableDenseNDimArray


from ..algebras import LieAlgebra


def commutator(A: Basic, B: Basic, anti=False) -> Basic:
    r"""Performs commutation brackets on A,B.

    If anti is False

    .. math::
        [ A, B ] = A * B - B * A

    Otherwise 

    .. math::
        \{A, B\} = A * B + B * A

    Args:
        A (Basic): Any mathematical object
        B (Basic): Any mathematical object
        anti (bool, optional): Anticommutation. Defaults to False.

    Returns:
        Basic: The (anti)commutation bracket result
    """
    if anti:
        return A * B + B * A
    return A * B - B * A


class Group(Basic):
    """The base class for all (lie) groups. The methods and properties
    in this class are basis independent and apply in a general sense. In
    order to write down the roots as matricies and vectors, we choose a 
    representation.
    """

    def __new__(cls, group: str, dim: int):
        obj = super().__new__(cls, sympify(dim))
        obj._group = group
        return obj

    @property
    def group(self) -> str:
        """Group type
        """
        return self._group

    @property
    def dimension(self) -> int:
        """Group dimension
        """
        return self.args[0]

    def generators(self) -> list:
        """Generalized matrix generators of the group.

        Abstract
        """
        pass

    def product(self, *args, **kwargs) -> list:
        """Calculates the product between several group representations.
        If backing algebra is a Lie Algebra, defaults to `LieAlgebra.tensor_product_decomposition`.

        Abstract
        """
        pass

    def sym_product(self, *args, as_tuple=False, **kwargs) -> list:
        """Calculates the product between several group representations passed by the symbolic name.
        If backing algebra is a Lie Algebra, defaults to `LieAlgebra.tensor_product_decomposition`

        Abstract
        """
        pass

    def conjugate(self, rep, symbolic=False):
        """Returns the conjugate representation. If the incoming rep is symbolic/named then it will return as such.

        Abstract
        """

    def irrep_lookup(self, irrep):
        """Returns the mathematical representation to the common name. Example would be returning `3` in SU(3) as `Matrix([[1,0]])`

        Abstract
        """


class LieGroup(Group):
    """Group that has a Lie Algebra associated with it.
    """

    def __init__(self, *args, **kwargs):
        """Used to set lazy properties
        """
        self._algebra = None
        self._structure_constants = (None, None)

    @property
    def algebra(self) -> LieAlgebra:
        """Backing Lie Algebra
        """
        return self._algebra

    def product(self, *args, **kwargs) -> list['Matrix']:
        """Uses tensor product decomposition to find the products between the 
        representations. Supported kwargs can be found on `LieAlgebra.tensor_product_decomposition`

        Returns:
            list[Matrix]: Tensor sum of the product.

        Examples
        =========
        >>> from liesym import SO
        >>> from sympy import Matrix
        >>> so10 = SO(10)
        >>> so10.product(Matrix([[1,0,0,0,0]]),Matrix([[1,0,0,0,0]]))
        [Matrix([[0, 0, 0, 0, 0]]), Matrix([[0, 1, 0, 0, 0]]), Matrix([[2, 0, 0, 0, 0]])]
        """

        # type: ignore
        return self.algebra.tensor_product_decomposition(args, **kwargs)

    def sym_product(self, *args, as_tuple=False, **kwargs) -> list[Union[Symbol, Tuple[Matrix, Symbol]]]:
        r"""Uses tensor product decomposition to find the products between the 
        representations. Supported kwargs can be found on `LieAlgebra.tensor_product_decomposition`.

        Args:
            as_tuple (bool, optional): If True, returns tuple of matrix rep with symbol. Defaults to False.

        Returns:
            list[Union[Symbol, Tuple[Matrix, Symbol]]]: list of symbols or list of (Matrix, Symbol)

        Examples
        ========
        >>> from liesym import SO
        >>> so10 = SO(10)
        >>> so10.sym_product("10","45")        
        [120, 10, 320]
        >>> so10.sym_product("10","45", as_tuple=True)
        [(Matrix([[0, 0, 1, 0, 0]]), 120), (Matrix([[1, 0, 0, 0, 0]]), 10), (Matrix([[1, 1, 0, 0, 0]]), 320)]
        >>> from liesym import SU
        >>> su3 = SU(3)
        >>> su3.sym_product('3', r'\bar{3}')
        [1, 8]
        """
        mats = [self.algebra.irrep_lookup(x) for x in args]

        results = []
        for rep in self.algebra.tensor_product_decomposition(mats, **kwargs):
            symbol = self.algebra.dim_name(rep)
            if as_tuple:
                results.append((rep, symbol))
            else:
                results.append(symbol)
        return results

    def conjugate(self, rep, symbolic=False):
        r"""Uses the underlying algebra to find the conjugate representation.

        Examples
        =========
        >>> from liesym import SU
        >>> from sympy import Matrix
        >>> su3 = SU(3)
        >>> su3.conjugate(Matrix([[1,0]]))
        Matrix([[0, 1]])
        >>> su3.conjugate(3, symbolic=True) # sympy prints without quotes
        \bar{3}
        """
        if symbolic:
            rep = str(rep)
            math_rep = self.algebra.irrep_lookup(rep)
            conj_rep = self.algebra.conjugate(math_rep)
            return self.algebra.dim_name(conj_rep)
        return self.algebra.conjugate(rep)

    def irrep_lookup(self, irrep: str) -> Matrix:
        """Uses the underlying algebra to do a lookup on the common name to find the matrix representation."""
        return self.algebra.irrep_lookup(irrep)

    def structure_constants(self, *idxs: int) -> Union[Basic, Matrix]:
        r"""Returns the structure constants of the group. Indexes start at 0 and constants maybe
        in different orderings than existing literature, but will still be in the Gell-Mann basis.
        Structure constants $f_{abc}$ is defined as

        .. math::
            [T_a, T_b] = i f_{abc} T_c

        where the group generators are $T$.

        Args:
            idxs (int): Optional postional arguments of 3 indices to return structure constant. If omitted, will return array.

        Returns:
            Union[Basic, Matrix]: If indicies are passed in, will return corresponding
            structure constant. Otherwise returns the 3dArry for entire group.
        """

        if self._structure_constants == (None, None):
            self._structure_constants = self._calculate_structure_constants()

        if len(idxs) > 0:
            if len(idxs) != 3:
                raise ValueError(
                    "3 indices need to be passed in if calling `structure_constants` with indices")
            else:
                [i, j, k] = idxs
                return self._structure_constants[0][i, j, k]
        return self._structure_constants[0]

    def _calculate_structure_constants(self):
        """Calculates the structure constants"""

        # For improved performance use explicit Symengine backend
        gens = [_CMatrix(x) for x in self.generators()]
        n = len(gens)

        f = MutableDenseNDimArray.zeros(n, n, n) * sympify("0")
        d = MutableDenseNDimArray.zeros(n, n, n) * sympify("0")

        for i in range(n):
            for j in range(n):
                for k in range(n):
                    f[i, j, k] = -2 * I * \
                        trace(commutator(gens[i], gens[j]) * gens[k])
                    d[i, j, k] = 2 * \
                        trace(commutator(
                            gens[i], gens[j], anti=True) * gens[k])
        return (f, d)

    def d_coeffecients(self, *idxs: int) -> Union[Basic, Matrix]:
        r"""Returns the d constants of the group. Indexes start at 0 and constants maybe
        in different orderings than existing literature, but will still be in the Gell-Mann basis.
        d constants $d_{abc}$ is defined as

        .. math::
            {T_a, T_b} = \frac{1}{N}\delta_{ab} + d_{abc} T_c

        where the group generators are $T$.

        Args:
            idxs (int): Optional postional arguments of 3 indices to return structure constant. If omitted, will return array.

        Returns:
            Union[Basic, Matrix]: If indicies are passed in, will return corresponding
            d constant. Otherwise returns the 3dArry for entire group.
        """
        if self._structure_constants == (None, None):
            self._structure_constants = self._calculate_structure_constants()

        if len(idxs) > 0:
            if len(idxs) != 3:
                raise ValueError(
                    "3 indices need to be passed in if calling `structure_constants` with indices")
            else:
                [i, j, k] = idxs
                return self._structure_constants[1][i, j, k]
        return self._structure_constants[1]

    def dynkin_index(self, irrep: Union[Basic, Matrix, str, int] = None, **kwargs) -> Basic:
        """Returns the dykin index for the arbitrary irreducible representation. This method 
        extends the underlying algebra method by allowing symbolic dim names to be passed.

        Args:
            irrep (Union[Basic, Matrix, str, int], optional): If None is passed, will default 
            to adjoint rep. Defaults to None.
            kwargs: Pass through to LieAlgebra.dynkin_index

        Returns:
            Basic: The irrep's dynkin index.
        """
        if irrep is None:
            return self.algebra.dynkin_index()
        elif isinstance(irrep, Matrix):
            return self.algebra.dynkin_index(irrep, **kwargs)
        elif isinstance(irrep, (Basic, str, int)):
            irrep = self.algebra.irrep_lookup(str(irrep))
            return self.algebra.dynkin_index(irrep)
        else:
            raise TypeError(
                "Only sympy basic types and sympy.Matrix are allowed.")

    def quadratic_casimir(self, irrep: Union[Basic, Matrix, str, int] = None, **kwargs) -> Basic:
        """Returns the quadratic casimir for the arbitrary irreducible representation. This method extends the underlying algebra method by allowing
        symbolic dim names to be passed.

        Args:
            irrep (Union[Basic, Matrix, str, int], optional): If None is passed, will default to adjoint rep. Defaults to None.
            kwargs: Pass through to LieAlgebra.quadratic_casimir

        Returns:
            Basic: The irrep's quadratic casimir.
        """
        if irrep is None:
            return self.algebra.quadratic_casimir()
        elif isinstance(irrep, Matrix):
            return self.algebra.quadratic_casimir(irrep, **kwargs)
        elif isinstance(irrep, (Basic, str, int)):
            irrep = self.algebra.irrep_lookup(str(irrep))
            return self.algebra.quadratic_casimir(irrep)
        else:
            raise TypeError(
                "Only sympy basic types and sympy.Matrix are allowed.")
